#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "thread/h/thread.h"

#define NSWITCHES 10000

typedef struct {
    int    i;
    double d;
    char   c;
    void*  p;
}   Tsd;

static thread_key_t tsd_key;

static void
myfree(void* ptr) {
    thr_do_trace("destructor called on ptr %p", ptr);
    (void) free(ptr);
    return;
}

static void
bar(void) {
    thread_t me = thr_self();
    Tsd*     ptr;


    if (thr_getspecific(tsd_key, (void **) &ptr) != THR_OKAY) {
        thr_perror("thr_getspecific");
        return;
    }

    if (ptr->i != (int)    me) { thr_do_trace("i field is %d", ptr->i); }
    if (ptr->d != (double) me) { thr_do_trace("d field is %g", ptr->d); }
    if (ptr->c != (char)   me) { thr_do_trace("c field is %d", ptr->c); }
    if (ptr->p != (void *) me) { thr_do_trace("p field is %p", ptr->p); }

    return;
}

static void*
foo(void* arg)
{
    int      i;
    Tsd*     ptr;
    thread_t me = thr_self();

    if ((ptr = (Tsd *) malloc(sizeof(Tsd))) == 0) {
        perror("malloc");
        return 0;
    }

    ptr->i = (int)    me;
    ptr->d = (double) me;
    ptr->c = (char)   me;
    ptr->p = (void *) me;

    if (thr_setspecific(tsd_key, ptr) != THR_OKAY) {
        thr_perror("thr_setspecific");
        return 0;
    }
    thr_do_trace("installed data, yielding to others");

    for (i = 0; i < NSWITCHES; i++) {
        thr_yield();
    }

    bar();
    thr_do_trace("finishing, verify destructor on %p", ptr);
    return 0;
}


int
main(void)
{
    unsigned i;

    thr_do_trace("demostration of thread-specific-data");

    if (thr_keycreate(&tsd_key, myfree) != THR_OKAY) {
        thr_perror("thr_keycreate");
        return -1;
    }

    for (i = 0; i < 10; i++) {
        thr_create(0, 0, foo, 0, 0, 0);
    }

    for (i = 0; i < 10; i++) {
        thr_join(0, 0, 0);
    }

    return 0;
}
